"""
Copyright (c) 2014-2017, Bitnine Inc.
Copyright (c) 2022, Vinzenz Feenstra

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
from typing import Optional

from psycopg import abc
from psycopg.adapt import Dumper, Loader
from psycopg.errors import InterfaceError

from psycopg3_agensgraph._edge import EdgeLoader
from psycopg3_agensgraph._oids import _EDGE_OID, _GRAPHPATH_OID, _VERTEX_OID
from psycopg3_agensgraph._vertex import VertexLoader


class Path(object):
    def __init__(self, vertices, edges):
        self.vertices = vertices
        self.edges = edges

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            return self.vertices == other.vertices and self.edges == other.edges
        return False

    def __len__(self):
        return len(self.edges)

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self)

    def __str__(self):
        p = [None] * (len(self.vertices) + len(self.edges))
        p[::2] = [str(v) for v in self.vertices]
        p[1::2] = [str(e) for e in self.edges]
        return "[%s]" % ",".join(p)


class PathDumper(Dumper):
    oid = _GRAPHPATH_OID

    def dump(self, value):
        return str(value).encode()


def _consume_path_element(buffer, start, nodes):
    loc = start
    size = len(buffer)
    while buffer[loc] != "{" and loc < size:
        loc += 1
    if buffer[loc] != "{" or loc >= size:
        raise InterfaceError(
            "bad graphpath data: '{}' position: {}".format(buffer, loc)
        )
    decoder = json.decoder.JSONDecoder()
    _, loc = decoder.raw_decode(buffer, loc)
    nodes.append(buffer[start:loc].strip(","))
    return loc


class PathLoader(Loader):
    def __init__(
        self,
        oid: int = _GRAPHPATH_OID,
        context: Optional[abc.AdaptContext] = None,
    ):
        super().__init__(oid, context)

    def load(self, value):
        if value is None:
            return None

        if not isinstance(value, str):
            value = bytes(value).decode()

        # ignore wrapping '[' and ']' characters
        pos = 1
        length = len(value) - 1

        vtokens = []
        etokens = []
        ve_tokens = vtokens, etokens
        while pos < length:
            pos = _consume_path_element(value, pos, ve_tokens[0])
            ve_tokens = ve_tokens[1], ve_tokens[0]

        vloader = VertexLoader(_VERTEX_OID)
        vertices = [vloader.load(t) for t in vtokens]
        eloader = EdgeLoader(_EDGE_OID)
        edges = [eloader.load(t) for t in etokens]
        return Path(vertices, edges)
