"""
Copyright (c) 2014-2018, Bitnine Inc.
Copyright (c) 2022, Vinzenz Feenstra

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from psycopg.adapt import Buffer, Dumper, pq
from psycopg.types.json import json

_ENCODERS = {}


def auto_encoder(v):
    encoder = _ENCODERS.get(type(v), None)
    if not encoder:
        raise ValueError(f"Do not know how to encode {type(v)} for postgresql")
    return encoder(v)


def entity_encoder(v):
    return f"{v}"


def iter_encoder(v):
    return f'[{",".join([auto_encoder(e) for e in v])}]'


def str_encoder(v):
    v = json.dumps(v)[1:-1]
    return "'" + v.replace("'", "''") + "'"


def key_encoder(k):
    if isinstance(k, str):
        return auto_encoder(k)
    return f"'{auto_encoder(k)}'"


def dct_encoder(v):
    return (
        f'{{{",".join([f"{key_encoder(k)}:{auto_encoder(v)}" for k, v in v.items()])}}}'
    )


def binary_encoder(v):
    raise ValueError("Bytes not supported")


_ENCODERS.update(
    {
        int: entity_encoder,
        list: iter_encoder,
        tuple: iter_encoder,
        set: iter_encoder,
        dict: dct_encoder,
        str: str_encoder,
        bytes: binary_encoder,
        float: entity_encoder,
        type(None): lambda _: "null",
        bool: lambda v: "true" if v else "false",
    }
)


class Property(object):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return auto_encoder(self.value)


class PropertyDumper(Dumper):
    format: pq.Format = pq.Format.TEXT

    def dump(self, value: Property) -> Buffer:
        return json.dumps(value.value, separators=(",", ":")).encode()
