"""
Copyright (c) 2014-2017, Bitnine Inc.
Copyright (c) 2022, Vinzenz Feenstra

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import re
from typing import Optional

from psycopg import abc
from psycopg.adapt import Dumper, Loader
from psycopg.errors import InterfaceError
from psycopg.types.json import json

from psycopg3_agensgraph._graphid import GraphIdLoader
from psycopg3_agensgraph._oids import _GRAPHID_OID, _VERTEX_OID
from psycopg3_agensgraph._property import Property

_pattern = re.compile(rb"(.+?)\[(.+?)\](.*)", re.S)


class Vertex(object):
    def __init__(self, label, vid, props):
        self.label = label if isinstance(label, str) else label.decode()
        self.vid = vid
        self.props = props

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            return self.vid == other.vid
        return False

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self)

    def __str__(self):
        return "%s[%s]%s" % (self.label, self.vid, Property(self.props))


class VertexDumper(Dumper):
    oid = _VERTEX_OID

    def dump(self, value: Vertex):
        return str(value).encode()


class VertexLoader(Loader):
    def __init__(
        self,
        oid: int = _VERTEX_OID,
        context: Optional[abc.AdaptContext] = None,
    ):
        super().__init__(oid, context)

    def load(self, value):
        if value is None:
            return None
        value = value if isinstance(value, (bytes, memoryview)) else value.encode()
        m = _pattern.match(value)
        if m:
            label = m.group(1)
            gloader = GraphIdLoader(_GRAPHID_OID)
            vid = gloader.load(m.group(2))
            props = json.loads(m.group(3))
            return Vertex(label, vid, props)
        else:
            raise InterfaceError("bad vertex representation: %s" % value)
